// $Id: main.c 12 2007-03-03 20:23:17Z salival $
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "uart.h"
#include "shiftregister.h"
#include "spi.h"
#include "cube.h"

#define UART_STATE_IDLE                 0
#define UART_STATE_FRAME_BYTEWISE       1
#define UART_STATE_FRAME_NIBBLEWISE     2

#define UART_COMMAND_CLEAR              'a'
#define UART_COMMAND_FRAME_BYTEWISE     'b'
#define UART_COMMAND_FRAME_NIBBLEWISE   'c'

volatile uint8_t step_flag;

static void test_pattern(void);
static void timer_init(void);
static void uart_fsm(void);

SIGNAL(TIMER0_OVF_vect)
{
    step_flag = 1;
}

int main(void)
{
    uint8_t* step_pointer;

    spi_init();
    cube_init();
    uart_init();
    timer_init();
    sei();

    step_pointer = cube_get_step_pointer();

    test_pattern();

    for(;;)
    {
        if (step_flag)
        {
            step_flag = 0;
            shiftregister_send(CUBE_SHIFTREGISTER_N, step_pointer);
            cube_calc_next_step();
        }
        if ( uart_bytes_received() )
        {
            uart_fsm();
        }
    }
    return 0;
}

static void test_pattern(void)
{
    uint8_t i;
    for (i=0; i<CUBE_SIZE; i++)
    {
        cube_set(i,0,0,1);
        cube_set(0,i,0,1);
        cube_set(0,0,i,1);
        cube_set(i,i,0,2);
        cube_set(i,0,i,5);
        cube_set(0,i,i,8);
        cube_set(i,i,i,15);
    }
}

//8bit-overflow clk/8 -> 3906.25Hz @ F_CPU = 8MHz
static void timer_init(void)
{
    TCCR0 |= (1<<CS01);
    TIMSK |= (1<<TOIE0);
}

static void uart_fsm(void)
{
    static uint8_t state = UART_STATE_IDLE;
    static uint8_t high_nibble;
    static uint8_t* pointer;
    static uint16_t counter;
    uint8_t c = uart_getc();

    switch(state)
    {
        case UART_STATE_IDLE:
            switch (c)
            {
                case UART_COMMAND_CLEAR:
                    cube_clear();
                    break;
                case UART_COMMAND_FRAME_BYTEWISE:
                    counter = CUBE_DISPLAY_RAM_SIZE;
                    pointer = cube_get_display_pointer();
                    state = UART_STATE_FRAME_BYTEWISE;
                    break;
                case UART_COMMAND_FRAME_NIBBLEWISE:
                    counter = CUBE_LED_TOTAL_N;
                    pointer = cube_get_display_pointer();
                    high_nibble = 0;
                    state = UART_STATE_FRAME_NIBBLEWISE;
                    break;
                default:
                    break;
            }
            break;
        case UART_STATE_FRAME_BYTEWISE:
            *(pointer++) = c;
            if (!--counter)
                state = UART_STATE_IDLE;
            break;
        case UART_STATE_FRAME_NIBBLEWISE:
            if (high_nibble)
            {
                *(pointer++) |= (c<<4);
            }
            else
            {
                *pointer = c & 0x0f;
            }

            high_nibble ^= 1;

            if (!--counter)
                state = UART_STATE_IDLE;
            break;
        default:
            break;
    }
}

