// $Id: uart.c 11 2007-03-03 18:54:29Z salival $

#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "uart.h"

#if UART_USE_PUTS_P
#include <avr/pgmspace.h>
#endif //UART_USE_PUTS_P

//hardware specific settings
#ifdef __AVR_ATmega8__
#   define _UART_RX_VECTOR                 USART_RXC_vect
#   define _UART_TX_VECTOR                 USART_UDRE_vect
#   define _UART_BAUD_H_REGISTER           UBRRH
#   define _UART_BAUD_L_REGISTER           UBRRL
#   define _UART_DATA_REGISTER             UDR
#   define _uart_enable_double_speed()     UCSRA |= (1<<U2X)
#   define _uart_hw_init()                 UCSRB = (UART_ENABLE_RX<<RXEN) | (UART_ENABLE_TX<<TXEN) | (UART_ENABLE_RX<<RXCIE)
#   define _uart_enable_tx_interrupt()     UCSRB |= (1<<UDRIE)
#   define _uart_disable_tx_interrupt()    UCSRB &= ~(1<<UDRIE)
#else
#   error
#endif //__AVR_xxx

#if UART_ENABLE_TX
static volatile struct {
    uint8_t write_index;
    uint8_t read_index;
    uint8_t count;
    uint8_t buffer[UART_TX_BUFFERSIZE];
} uart_tx_fifo;
#endif //UART_ENABLE_TX

//receive buffer
#if UART_ENABLE_RX
static volatile struct {
    uint8_t write_index;
    uint8_t read_index;
    uint8_t count;
    uint8_t overflow;
    uint8_t buffer[UART_RX_BUFFERSIZE];
} uart_rx_fifo;
#endif //UART_ENABLE_RX

void uart_init(void)
{
#if UART_ENABLE_RX
    uart_rx_fifo.read_index = 0;
    uart_rx_fifo.write_index = 0;
    uart_rx_fifo.count = 0;
    uart_rx_fifo.overflow = 0;
#endif //UART_ENABLE_RX

#if UART_ENABLE_TX
    uart_tx_fifo.read_index = 0;
    uart_tx_fifo.write_index = 0;
    uart_tx_fifo.count = 0;
#endif //UART_ENABLE_TX

#if UART_DOUBLE_SPEED
    _UART_BAUD_H_REGISTER = (uint8_t) ( (F_CPU/8/UART_BAUDRATE-1)>>8 );
    _UART_BAUD_L_REGISTER = (uint8_t) (F_CPU/8/UART_BAUDRATE-1);
    _uart_enable_double_speed();
#else
    _UART_BAUD_H_REGISTER = (uint8_t) ( (F_CPU/16/UART_BAUDRATE-1)>>8 );
    _UART_BAUD_L_REGISTER = (uint8_t) (F_CPU/16/UART_BAUDRATE-1);
#endif //UART_DOUBLE_SPEED

    _uart_hw_init();
}

#if UART_ENABLE_RX
uint8_t uart_overflow(void)
{
    return uart_rx_fifo.overflow;
}

uint8_t uart_bytes_received(void)
{
    return uart_rx_fifo.count;
}

uint8_t uart_getc (void)
{
    register uint8_t index = uart_rx_fifo.read_index;
    register uint8_t c;
    register uint8_t sreg;

    //block until data available
    while(!uart_rx_fifo.count);

    c = uart_rx_fifo.buffer[index];

    sreg = SREG;
    cli();
    uart_rx_fifo.count--;
    SREG = sreg;

    if(index < UART_RX_BUFFERSIZE-1 )
        uart_rx_fifo.read_index = ++index;
    else
        uart_rx_fifo.read_index = 0;

    return c;
}

SIGNAL(_UART_RX_VECTOR)
{
    register uint8_t count = uart_rx_fifo.count;
    register uint8_t index;

    if( count < UART_RX_BUFFERSIZE-1 )
    {
        index = uart_rx_fifo.write_index;
        uart_rx_fifo.buffer[index] = _UART_DATA_REGISTER;
        uart_rx_fifo.count = ++count;

        if( index < UART_RX_BUFFERSIZE-1 )
            uart_rx_fifo.write_index = ++index;
        else
            uart_rx_fifo.write_index = 0;
    }
    else
        uart_rx_fifo.overflow = 1;
}
#endif //UART_ENABLE_RX

#if UART_ENABLE_TX
void uart_putc (uint8_t c)
{
    register uint8_t index = uart_tx_fifo.write_index;
    register uint8_t sreg;

    //block if buffer is full
    while( uart_tx_fifo.count >= UART_TX_BUFFERSIZE ) ;

    uart_tx_fifo.buffer[index] = c;

    sreg = SREG;
    cli();
    uart_tx_fifo.count++;
    _uart_enable_tx_interrupt();
    SREG = sreg;

    if( index < UART_TX_BUFFERSIZE-1 )
        uart_tx_fifo.write_index = ++index;
    else
        uart_tx_fifo.write_index = 0;
}

#if UART_USE_PUTS
void uart_puts(const uint8_t* s)
{
    while (*s)
        uart_putc( *(s++) );
}
#endif //UART_USE_PUTS

#if UART_USE_PUTS_P
void uart_puts_p(const uint8_t* s)
{
    register uint8_t c;
    while( (c = pgm_read_byte(s++)) )
        uart_putc(c);
}
#endif //UART_USE_PUTS_P

ISR(_UART_TX_VECTOR)
{
    register uint8_t index = uart_tx_fifo.read_index;
    register uint8_t count = uart_tx_fifo.count;

    _UART_DATA_REGISTER = uart_tx_fifo.buffer[index];
    uart_tx_fifo.count = --count;

    if(!count)
        _uart_disable_tx_interrupt();

    if( index < UART_TX_BUFFERSIZE-1 )
        uart_tx_fifo.read_index = ++index;
    else
        uart_tx_fifo.read_index = 0;
}
#endif //UART_ENABLE_TX

